// A separate-chaining hash table.

#pragma once

#include "Dyn_array.hpp"
#include <memory>

template <class K, class V>
class Hash_map
{
public:
    explicit Hash_map(size_t initial_capacity = 8);

    size_t size() const;

    V* find(K const&);
    V const* find(K const&) const;

    V& operator[](K const&);
    void insert(K const&, V const&);
    void remove(K const&);

    void get_keys(Dyn_array<K const*>&) const;

private:
    struct cons_;

    using link_ = std::shared_ptr<cons_>;

    struct cons_
    {
        K key;
        V val;
        link_ next;

        cons_(K const& k, V const& v, link_ n)
                : key(k), val(v), next(std::move(n))
        { }
    };

    size_t size_;
    size_t capacity_;
    std::unique_ptr<link_[]> buckets_;
    std::hash<K> hasher_;

    // private helpers
    static V* find_in_list_(K const&, cons_*);
    void cons_bucket_(size_t bucket, K const&, V const&);
};

template<class K, class V>
Hash_map<K, V>::Hash_map(size_t initial_capacity)
        : size_(0)
        , capacity_(initial_capacity)
        , buckets_(std::make_unique<link_[]>(capacity_))
{ }

template<class K, class V>
size_t Hash_map<K, V>::size() const
{
    return size_;
}


template<class K, class V>
V* Hash_map<K, V>::find(K const& key)
{
    size_t index = hasher_(key) % capacity_;
    return find_in_list_(key, buckets_[index].get());
}

template<class K, class V>
V const* Hash_map<K, V>::find(K const& key) const
{
    size_t index = hasher_(key) % capacity_;
    return find_in_list_(key, buckets_[index].get());
}

template<class K, class V>
V& Hash_map<K, V>::operator[](K const& key)
{
    size_t index = hasher_(key) % capacity_;
    V* result = find_in_list_(key, buckets_[index].get());

    if (result != nullptr) return *result;

    cons_bucket_(index, key, V());
    return buckets_[index]->val;
}

template<class K, class V>
void Hash_map<K, V>::insert(K const& key, V const& val)
{
    size_t index = hasher_(key) % capacity_;
    V* existing = find_in_list_(key, buckets_[index].get());

    if (existing != nullptr) {
        *existing = val;
        return;
    }

    cons_bucket_(index, key, val);
    ++size_;
}

template<class K, class V>
void Hash_map<K, V>::remove(K const& key)
{
    size_t index = hasher_(key) % capacity_;

    link_* curr = &buckets_[index];

    while (*curr != nullptr) {
        if ((*curr)->key == key) {
            *curr = std::move((*curr)->next);
            --size_;
            return;
        }
    }
}

template<class K, class V>
void Hash_map<K, V>::get_keys(Dyn_array<K const*>& da) const
{
    for (size_t i = 0; i < capacity_; ++i) {
        cons_* curr = buckets_[i].get();

        while (curr != nullptr) {
            da.push_back(&curr->key);
            curr = curr->next.get();
        }
    }
}

template<class K, class V>
V* Hash_map<K, V>::find_in_list_(K const& key, cons_* curr)
{
    while (curr != nullptr) {
        if (curr->key == key)
            return &curr->val;
        curr = curr->next.get();
    }

    return nullptr;
}

template<class K, class V>
void Hash_map<K, V>::cons_bucket_(size_t bucket, K const& key, V const& val)
{
    buckets_[bucket] = std::make_shared<cons_>(
            key, val, std::move(buckets_[bucket])
    );
}

