#pragma once

#include <stddef.h>

// Dynamic arrays of integers are represented as a pointer to an
// incomplete implementation type.
//
// This means that clients (.c files that include this .h file,
// such as int_da_client.c) know that there exists some int_da
// struct, but they don't get to know what it is and have to
// refer to it with a pointer.
typedef struct int_da* int_da_t;

// Allocates a new array with the given initial capacity.
//
// Returns NULL if memory cannot be allocated.
//
// The resulting array must be deallocated with `int_da_destroy`.
int_da_t int_da_create(size_t capacity);

// Deallocates the given array, recovering its storage.
//
// PRECONDITION: the given array is a non-NULL pointer to an object
// that has been returned by `int_da_create` and has not previously
// been passed to `int_da_destroy`.
void int_da_destroy(int_da_t);

// Returns the number of elements of the given array.
size_t int_da_size(int_da_t);

// Returns a pointer to the `i`th element of the given array.
//
// PRECONDITION: i is in bounds
//
// (We could have separate `int_da_get` and `int_da_set` operations,
// but this does both.)
int* int_da_index(int_da_t, size_t i);

// Increments the size of the array, returning a pointer to the
// new (uninitialized) element.
//
// If the array needs to grow and memory cannot be allocated, then
// the array is still valid and the returned pointer is NULL.
int* int_da_push_back(int_da_t);

// Removes the last element of the array, decrementing its size.
//
// PRECONDITION: the array is not empty
void int_da_pop_back(int_da_t);

